# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import lsst.pex.config as pexConfig


class ConfigTest(unittest.TestCase):
    def setUp(self):
        """Note: the classes are defined here in order to test the register
        decorator.
        """

        class ParentConfig(pexConfig.Config):
            pass

        self.registry = pexConfig.makeRegistry(doc="unit test configs", configBaseType=ParentConfig)

        class FooConfig1(ParentConfig):
            pass

        self.fooConfig1Class = FooConfig1

        class FooConfig2(ParentConfig):
            pass

        self.fooConfig2Class = FooConfig2

        class Config1(pexConfig.Config):
            pass

        self.config1Class = Config1

        class Config2(pexConfig.Config):
            pass

        self.config2Class = Config2

        @pexConfig.registerConfigurable("foo1", self.registry)
        class FooAlg1:
            ConfigClass = FooConfig1

            def __init__(self, config):
                self.config = config

            def foo(self):
                pass

        self.fooAlg1Class = FooAlg1

        class FooAlg2:
            ConfigClass = FooConfig2

            def __init__(self, config):
                self.config = config

            def foo(self):
                pass

        self.registry.register("foo2", FooAlg2, FooConfig2)
        self.fooAlg2Class = FooAlg2

        # override Foo2 with FooConfig1
        self.registry.register("foo21", FooAlg2, FooConfig1)

    def tearDown(self):
        del self.registry
        del self.fooConfig1Class
        del self.fooConfig2Class
        del self.fooAlg1Class
        del self.fooAlg2Class

    def testBasics(self):
        self.assertEqual(self.registry["foo1"], self.fooAlg1Class)
        self.assertEqual(self.registry["foo2"].ConfigClass, self.fooConfig2Class)
        self.assertEqual(self.registry["foo21"].ConfigClass, self.fooConfig1Class)

        self.assertEqual(set(self.registry.keys()), set(("foo1", "foo2", "foo21")))

    def testWrapper(self):
        wrapper21 = self.registry["foo21"]
        foo21 = wrapper21(wrapper21.ConfigClass())
        self.assertIsInstance(foo21, self.fooAlg2Class)

    def testReplace(self):
        """Test replacement in registry (should always fail)"""
        self.assertRaises(Exception, self.registry.register, "foo1", self.fooAlg2Class)
        self.assertEqual(self.registry["foo1"], self.fooAlg1Class)

    def testNesting(self):
        """Make sure nesting a config with a RegistryField doesn't deep-copy
        the registry."""

        class MidConfig(pexConfig.Config):
            field = self.registry.makeField("docs for registry field")

        class TopConfig(pexConfig.Config):
            middle = pexConfig.ConfigField(dtype=MidConfig, doc="docs for middle")

        self.assertIs(MidConfig.field.registry, self.registry)
        middle = MidConfig()
        top = TopConfig()
        self.assertIs(middle.field.registry, self.registry)
        self.assertIs(top.middle.field.registry, self.registry)

    def testRegistryField(self):
        class C1(pexConfig.Config):
            r = self.registry.makeField("registry field")

        for t in C1.r.typemap:
            self.assertEqual(C1.r.typemap[t], self.registry[t].ConfigClass)

        c = C1()
        c.r = "foo2"
        c.r.apply()

    def testExceptions(self):
        class C1(pexConfig.Config):
            r = self.registry.makeField("registry field", multi=True, default=[])

        c = C1()

        def fail(name):  # lambda doesn't like |=
            c.r.names |= [name]

        self.assertRaises(pexConfig.FieldValidationError, fail, "bar")


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import lsst.pex.config as pexConfig
import lsst.pex.config.history as pexConfigHistory


class PexTestConfig(pexConfig.Config):
    a = pexConfig.Field("Parameter A", float, default=1.0)


class HistoryTest(unittest.TestCase):
    def testHistory(self):
        b = PexTestConfig()
        b.update(a=4.0)
        pexConfigHistory.Color.colorize(False)
        output = b.formatHistory("a", writeSourceLine=False)

        # The history differs depending on how the tests are executed and might
        # depend on pytest internals. We therefore test the output for the
        # presence of strings that we know should be there.

        # For reference, this is the output from running with unittest.main()
        """a
1.0 unittest.main()
    self.runTests()
    self.result = testRunner.run(self.test)
    test(result)
    return self.run(*args, **kwds)
    test(result)
    return self.run(*args, **kwds)
    testMethod()
    b = PexTestConfig()
    a = pexConfig.Field('Parameter A', float, default=1.0)
4.0 unittest.main()
    self.runTests()
    self.result = testRunner.run(self.test)
    test(result)
    return self.run(*args, **kwds)
    test(result)
    return self.run(*args, **kwds)
    testMethod()
    b.update(a=4.0)"""

        self.assertTrue(output.startswith("a\n1.0"))
        print(output)
        self.assertIn(
            """
    b = PexTestConfig()
    a = pexConfig.Field("Parameter A", float, default=1.0)
4.0""",
            output,
        )

        self.assertIn("\n    b.update(a=4.0)", output)


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

import lsst.pex.config as pexConfig


class Config1(pexConfig.Config):
    f = pexConfig.Field("f", float, default=3.0)

    def _collectImports(self):
        # Exists to test that imports of dict values are collected
        self._imports.add("builtins")


class Config2(pexConfig.Config):
    d1 = pexConfig.ConfigDictField("d1", keytype=str, itemtype=Config1, itemCheck=lambda x: x.f > 0)


class Config3(pexConfig.Config):
    field1 = pexConfig.ConfigDictField(keytype=str, itemtype=pexConfig.Config, default={}, doc="doc")


class ConfigDictFieldTest(unittest.TestCase):
    def testConstructor(self):
        try:

            class BadKeytype(pexConfig.Config):
                d = pexConfig.ConfigDictField("...", keytype=list, itemtype=Config1)

        except Exception:
            pass
        else:
            raise SyntaxError("Unsupported keytypes should not be allowed")

        try:

            class BadItemtype(pexConfig.Config):
                d = pexConfig.ConfigDictField("...", keytype=int, itemtype=dict)

        except Exception:
            pass
        else:
            raise SyntaxError("Unsupported itemtypes should not be allowed")

        try:

            class BadItemCheck(pexConfig.Config):
                d = pexConfig.ConfigDictField("...", keytype=str, itemtype=Config1, itemCheck=4)

        except Exception:
            pass
        else:
            raise SyntaxError("Non-callable itemCheck should not be allowed")

        try:

            class BadDictCheck(pexConfig.Config):
                d = pexConfig.DictField("...", keytype=int, itemtype=Config1, dictCheck=4)

        except Exception:
            pass
        else:
            raise SyntaxError("Non-callable dictCheck should not be allowed")

    def testAssignment(self):
        c = Config2()
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", {3: 3})
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", {"a": 0})
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", [1.2, 3, 4])
        c.d1 = None
        c.d1 = {"a": Config1, "b": Config1()}

    def testValidate(self):
        c = Config2()
        self.assertRaises(pexConfig.FieldValidationError, Config2.validate, c)

        c.d1 = {"a": Config1(f=0)}
        self.assertRaises(pexConfig.FieldValidationError, Config2.validate, c)

        c.d1["a"].f = 5
        c.validate()

    def testInPlaceModification(self):
        c = Config2(d1={})
        self.assertRaises(pexConfig.FieldValidationError, c.d1.__setitem__, 1, 0)
        self.assertRaises(pexConfig.FieldValidationError, c.d1.__setitem__, "a", 0)
        c.d1["a"] = Config1(f=4)
        self.assertEqual(c.d1["a"].f, 4)

    def testSave(self):
        c = Config2(d1={"a": Config1(f=4)})
        c.save("configDictTest.py")

        # verify _collectImports is called on all the configDictValues
        stringOutput = c.saveToString()
        self.assertIn("import builtins", stringOutput)

        rt = Config2()
        rt.load("configDictTest.py")

        os.remove("configDictTest.py")
        self.assertEqual(rt.d1["a"].f, c.d1["a"].f)

        c = Config2()
        c.save("emptyConfigDictTest.py")
        rt.load("emptyConfigDictTest.py")
        os.remove("emptyConfigDictTest.py")

        self.assertIsNone(rt.d1)

    def testToDict(self):
        c = Config2(d1={"a": Config1(f=4), "b": Config1})
        dict_ = c.toDict()
        self.assertEqual(dict_, {"d1": {"a": {"f": 4.0}, "b": {"f": 3.0}}})

    def testFreeze(self):
        c = Config2(d1={"a": Config1(f=4), "b": Config1})
        c.freeze()

        self.assertRaises(pexConfig.FieldValidationError, setattr, c.d1["a"], "f", 0)

    def testNoArbitraryAttributes(self):
        c = Config2(d1={})
        self.assertRaises(pexConfig.FieldValidationError, setattr, c.d1, "should", "fail")

    def testEquality(self):
        """Test ConfigDictField.__eq__

        We create two configs, with the keys explicitly added in a different
        order and test their equality.
        """
        keys1 = ["A", "B", "C"]
        keys2 = ["X", "Y", "Z", "a", "b", "c", "d", "e"]

        c1 = Config3()
        c1.field1 = {k: pexConfig.Config() for k in keys1}
        for k in keys2:
            c1.field1[k] = pexConfig.Config()

        c2 = Config3()
        for k in keys2 + keys1:
            c2.field1[k] = pexConfig.Config()

        self.assertTrue(pexConfig.compareConfigs("test", c1, c2))


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

try:
    import testLib
except ImportError:
    testLib = None

import pickle


@unittest.skipIf(testLib is None, "C++ tests disabled")
class WrapTest(unittest.TestCase):
    def testMakeControl(self):
        """Test making a C++ Control object from a Config object."""
        config = testLib.ConfigObject()
        config.foo = 2
        config.bar.append("baz")
        control = config.makeControl()
        self.assertTrue(testLib.checkControl(control, config.foo, config.bar.list()))

    def testReadControl(self):
        """Test reading the values from a C++ Control object into a Config
        object."""
        control = testLib.ControlObject()
        control.foo = 3
        control.bar = ["zot", "yox"]
        config = testLib.ConfigObject()
        config.readControl(control)
        self.assertTrue(testLib.checkControl(control, config.foo, config.bar.list()))

    def testDefaults(self):
        """Test that C++ Control object defaults are correctly used as defaults
        for Config objects."""
        config = testLib.ConfigObject()
        control = testLib.ControlObject()
        self.assertTrue(testLib.checkControl(control, config.foo, config.bar.list()))

    def testPickle(self):
        """Test that C++ Control object pickles correctly"""
        config = testLib.ConfigObject()
        new = pickle.loads(pickle.dumps(config))
        self.assertTrue(config.compare(new))
        self.assertTrue(new.compare(config))


@unittest.skipIf(testLib is None, "C++ tests disabled")
class NestedWrapTest(unittest.TestCase):
    def testMakeControl(self):
        """Test making a C++ Control object from a Config object."""
        config = testLib.OuterConfigObject()
        self.assertIsInstance(config.a, testLib.InnerConfigObject)
        config.a.p = 5.0
        config.a.q = 7
        config.b = 2
        control = config.makeControl()
        self.assertTrue(testLib.checkNestedControl(control, config.a.p, config.a.q, config.b))

    def testReadControl(self):
        """Test reading the values from a C++ Control object into a Config
        object."""
        control = testLib.OuterControlObject()
        control.a.p = 6.0
        control.a.q = 4
        control.b = 3
        config = testLib.OuterConfigObject()
        config.readControl(control)
        self.assertTrue(testLib.checkNestedControl(control, config.a.p, config.a.q, config.b))

    def testDefaults(self):
        """Test that C++ Control object defaults are correctly used as defaults
        for Config objects."""
        config = testLib.OuterConfigObject()
        control = testLib.OuterControlObject()
        self.assertTrue(testLib.checkNestedControl(control, config.a.p, config.a.q, config.b))

    def testInt64(self):
        """Test that we can wrap C++ Control objects with int64 members."""
        config = testLib.OuterConfigObject()
        control = testLib.OuterControlObject()
        self.assertTrue(testLib.checkNestedControl(control, config.a.p, config.a.q, config.b))
        self.assertGreater(config.a.q, 1 << 30)
        self.assertGreater(control.a.q, 1 << 30)


if __name__ == "__main__":
    unittest.main()
